#!/usr/bin/python3

import sys
import os
from consts import *
import options as Options


def process_output(page, data, indexes):

  if not Options.FindLinks: return data

  text = ''
  word = ''
  i = 0
  length = len(data)
  while i <= length:
    if i < length and (data[i].isalnum() or data[i] == '_'):
      word += data[i]
    else:
      if word != '':
        found = False

        # Check all indexes for this key
        for x in indexes[TYPE_MIN:TYPE_MAX]:
          if word in list(x.keys()) and x[word] != page:
            text += outlib.link(word, x[word])
            #if Options.Debug: print('link: %s -> %s' % (word, x[word]))
            found = True
            break

        if not found:
          #if Options.Debug: print('word: %s' % word)
          text += word

      word = ''
      if i < length: text += data[i]
      #if Options.Debug: print('text: %s' % text)

    i += 1

  return text


# for each entry in the dict 'data', create a file of content
# for each entry in indexes[], that does not have a content file, create a link file
def write_output(out, indexes, data, lib):

  # for each dict, corresponding to a type
#  for index in indexes[TYPE_MIN:TYPE_MAX]:
  for index_num in range(TYPE_MIN, TYPE_MAX):
    index = indexes[index_num]

    # for each page name
    keys = list(index.keys())
    keys.sort()
    for k in keys:
      #if Options.Debug: print('\n\nkey %s' % k)

      if k == index[k]:
        # if 'tag' is a key in data, this is a content file

        text = process_output(k, data[k], indexes)

        outlib.write_content(out, index[k], text, lib, index_num)

      else:
        # otherwise it's a stronghelp lookup link file
        outlib.write_sh_link(k, index[k])

  return


def find_fn_name(line):

  i = line.find('(')

  if i >= 0:
    j = line.rfind(' ', 0, i)
    name = line[j+1:i].strip('*')
    if name != '': return name

  return None


def find_def_name(line):

  i = line.find('#define')

  if i >= 0:
    j = line[i+8:].find(' ')
    name = line[i+8:j+i+8]
    if name != '': return name

  return None


def add_pages(pages, data, index, output):
  tag = pages[0]
  for p in pages:
    index[p] = tag		# name => tag (base page name is the tag)
    #if Options.Debug: print('link %s -> %s' % (p, tag))

  output[pages[0]] = data	# tag => page text

  #if Options.Debug: print('data %s -> %s\n' % (tag, data[:100]))

  return



# write any current pages
def write_current_page(pages, data, index, output):
  global State, PrevState

  if State == STATE_IN_TEXT or PrevState == STATE_IN_TEXT: data += outlib.end_text()
  elif State == STATE_IN_CODE or PrevState == STATE_IN_CODE: data += outlib.end_code()
  add_pages(pages, data, index, output)
  State = STATE_NOPAGE
  PrevState = STATE_NOPAGE
  if Options.Debug: print('  write_current_page fin: %s' % pages)
  return


def process_file(path, lib, indexes, output):

  global State
  global PrevState

  pages = []
  match = [ ('/**fn:',     '**fn*/',     TYPE_FN),
            ('/**def:',    '**def*/',    TYPE_DEF),
            ('/**pragma:', '**pragma*/', TYPE_PRAGMA),
            ('/**type:',   '**type*/',   TYPE_TYPEDEF),
            ('/**note:',   '**note*/',   TYPE_NOTE),
            ('/**index:',  '**index*/',  TYPE_INDEX),		## put text on the index page
          ]

#  lines = [line.strip() for line in file(path).readlines()]

  State = STATE_NOPAGE
  PrevState = STATE_NOPAGE
  type = TYPE_NONE
  pages_specified = False
  close = ''
  data = ''

#  for line in lines:

#  for line in file(path).readlines():
  with open(path, mode='r') as file:
    for line in file:
      # get next line, remove trailing whitespace
      line = line.rstrip()
      #if Options.Debug: print('state %d (%d) line %s' % (State, PrevState, line))
      if line == '': continue

      # look for a page start marker
      for (start, c, t) in match:
        i = line.find(start)
        if i >= 0:
          # found page start, write any current page[s]
          if pages != []: write_current_page(pages, data, indexes[type], output)
          pages = []
          data = ''
          close = c
          type = t
          if Options.Debug: print('  start page type %d: %s' % (type, line))
          break

      if i >= 0:
        # found start of page
        j = line[i:].find(':')		# skip start marker
        if j < 0:
          print('Bad start line: %s' % line)
          State = STATE_NOPAGE
          continue			# badly formatted line?

        # is the comment closed on this line?
        if '*/' in line: State = STATE_START_CODE
        else: State = STATE_START_TEXT

        # strip comment close, to leave list of page titles (possibly empty)
        line = line[i+j+1:].rstrip('*/')

        # build list of page titles - ie links to this data
        if type != TYPE_INDEX:
          for page in line.split():
            if Options.Debug: print('  specified page: %s' % page)
            pages.append(page)
        else:
          pages.append('__index')

        if pages != []: pages_specified = True
        else: pages_specified = False
        continue			# get next line

      # didn't find start of new page, check for check for end of current page
      if State != STATE_NOPAGE:
        if close in line or '**/' in line:
          if pages != []: write_current_page(pages, data, indexes[type], output)
          pages = []
          data = ''
          type = TYPE_NONE
          close = ''
          continue					# get next line

      # decide whether this line starts a comment
      if State != STATE_NOPAGE:
        pline = line.strip()
        if len(pline) >= 2 and pline[0:2] == '/*': State = STATE_START_TEXT

      if State in (STATE_START_CODE, STATE_IN_CODE) and not pages_specified:
        # if we find something like a function name, add it to the list of page names
        # assuming no page names were specified in the start command
        if type == TYPE_FN:
          page = find_fn_name(line)
          if page != None and page not in pages:
            State = STATE_START_CODE
            pages.append(page)
            if Options.Debug: print('  found fn name: %s' % page)
        elif type == TYPE_DEF:
          page = find_def_name(line)
          if page != None and page not in pages:
            State = STATE_START_CODE
            pages.append(page)
            if Options.Debug: print('  found def name: %s' % page)


      pline = ''
      if State == STATE_START_TEXT or State == STATE_IN_TEXT:
        pline = outlib.process_text(line)
      elif State == STATE_START_CODE or State == STATE_IN_CODE:
        pline = outlib.process_code(line)

      if pline != '':
        if State == STATE_START_TEXT:
          if PrevState == STATE_IN_CODE: data += outlib.end_code()
          if PrevState != STATE_IN_TEXT: data += outlib.start_text()
          State = STATE_IN_TEXT

        if State == STATE_START_CODE:
          if PrevState == STATE_IN_TEXT: data += outlib.end_text()
          if PrevState != STATE_IN_CODE: data += outlib.start_code()
          State = STATE_IN_CODE

        data += pline
        PrevState = State

      if (State == STATE_START_TEXT or State == STATE_IN_TEXT) and '*/' in line: State = STATE_START_CODE

  return


# start of main code
# command line options:
# -source <dir>		the header files directory
# -output <dir>		where to write output files
# [-fullindex]		index all pages and links.
#			Otherwise just index main pages
# [-indexcols <n>]	use <n> columns on the index page
# [-debug]		be verbose
# [-html]


for i in range(len(sys.argv)):

  if sys.argv[i] == '-debug':
    Options.Debug = True

  elif sys.argv[i] == '-fullindex':
    Options.FullIndex = True

  elif sys.argv[i] == '-findlinks':
    Options.FindLinks = True

  elif sys.argv[i] == '-html':
    import html as outlib

  elif sys.argv[i] == '-indexcols':
    i += 1
    l = sys.argv[i].split(',')
    for a in range(len(l)):
      Options.IndexCols[a] = int(l[a])

  elif sys.argv[i] == '-source':
    i += 1
    Options.SourceDir = sys.argv[i]

  elif sys.argv[i] == '-output':
    i += 1
    Options.OutputDir = sys.argv[i]

  elif sys.argv[i] == '-root':
    i += 1
    Options.RootPage = sys.argv[i]

if Options.SourceDir == None or Options.OutputDir == None:
  print('Must specify source and output directories')
  sys.exit()

print('Debug: %s' % Options.Debug)
print('SourceDir: %s' % Options.SourceDir)
print('OutputDir: %s' % Options.OutputDir)
print('FullIndex: %s' % Options.FullIndex)
print('FindLinks: %s' % Options.FindLinks)
print('IndexCols: %s' % Options.IndexCols)
print('RootPage: %s' % Options.RootPage)

Libs = os.listdir(Options.SourceDir)
#if Options.Debug:
print('Input files: %s' % Libs)

try:
  outlib.loaded()
except NameError:
  import sh as outlib
  for dir in 'abcdefghijklmnopqrstuvwxyz_':
    os.mkdir('%s.[%s]' % (Options.OutputDir, dir))

for lib in Libs:
  indexes = ( {}, {}, {}, {}, {}, {} )	# name => tag
  data = {}				# tag => data

  print('file: %s ' % (Libs[Libs.index(lib)]), end='')

  process_file('%s.%s' % (Options.SourceDir, lib), lib, indexes, data)
  if data != {}:
    out = outlib.write_index(indexes, data, lib)	# opens file
    write_output(out, indexes, data, lib)

    if out != None:
      out.write(outlib.end_page())
      out.close()

    print('ok')

  else:
    # remove this entry from Libs
    Libs[Libs.index(lib)] = ''
    print('no content')

if Options.RootPage != None:
  outlib.write_root(Libs)

print('done')
